/* ---------------------------------------- */
#include <stdio.h>

/* ---------------------------------------- */
#define PRGNAME		"Rom Reader"
#define PRGVER		0x10
#define PRGAUTH		"RomBor 11/2021"

#define TRANS		0xDF70
#define TRANS2		0xDFB0
#define TRANS3		0xDFF0
#define TMEM		0xFFFF

#define BLKSIZE		1024
#define NOTSET		-1

#define EOK		0
#define EALLOC		-1
#define EOPTVAL		-2
#define EBADVAL		-3
#define EUNKOPT		-4
#define EOPTCOMB	-5
#define EMISOPT		-6
#define EBADFN		-7
#define EFILCRE		-8
#define EFILWRI		-9

/* ---------------------------------------- */
/* globalne premenne */
int bb, cb, sb, dump, nfn;
unsigned ab, lb;
char fn[12], *pFn, *pMem;

/* ---------------------------------------- */
int main(argc, argv)
  int argc;
  char **argv;
{
  char c, *p;
  char *pT;
  int err;

  /* priprava volania rutiny TRANSFER
     na adresu 0xDFB0 pre PMD 85-2A */
  pT = TRANS;
  initb(pT, "33,0,0,57,34,122,192,49,0,224,62,9,195,176,223");
  pT = TRANS3;
  initb(pT, "211,247,42,122,192,249,201");
  pT = TRANS2;
  initb(pT, "211,247,205,0,140,0,0,0,0,0,0,62,8,195,240,223");

  /* uprava rutiny pre PMD 85-3 */
  p = TMEM;
  c = 255 - *p;
  *p = c;
  if (*p != c) {
    *(pT + 4) = 236;
    *(pT + 12) = 9;
  }

  version();

  err = EOK;
  do {
    if (argc == 1) {
      usage();
      break;
    }

    pMem = alloc(BLKSIZE);
    if (!pMem) {
      err = prtErr(EALLOC, NULL, NULL);
      break;
    }

    err = procArgs(argc, argv);
    if (err != EOK)
      break;

    if (dump)
      dumpRom();
    else
      saveRom();

  } while (FALSE);

  if (pMem)
    free(pMem);

  return -err;
}
/* ---------------------------------------- */
void saveRom()
{
  unsigned adr, len, sep, l, s;
  FILE *f;

  if (ab != NOTSET) {
    adr = ab;
    len = lb;
    sep = 0xFFFF;
    printf("Adresa: %04x, Dlzka: %04x\n", adr, len);
  }
  else {
    adr = bb * BLKSIZE;
    len = cb * BLKSIZE;
    sep = (sb == NOTSET) ? 0xFFFF : (sb * BLKSIZE);
    printf("Od bloku: %d, Pocet: %d\n", bb, cb);
    if (sb != NOTSET)
      printf("Blokov na subor: %d\n", sb);
  }

  f = NULL;
  l = 0;
  s = sep;
  while (len) {
    if (!l)
      l = readRom(adr, len);

    if (!f) {
      if (sb != NOTSET) {
        fn[nfn]++;
        if (fn[nfn] == ':') {
          fn[nfn] = '0';
          fn[nfn - 1]++;
        }
      }
      f = fopen(fn, "wb");
      if (!f) {
        prtErr(EFILCRE, fn, NULL);
        break;
      }
      printf("Zapisujem subor: %s\n", fn);
    }

    if (fwrite(pMem, l, 1, f) != 1) {
      prtErr(EFILWRI, fn, NULL);
      break;
    }

    adr += l;
    len -= l;
    s -= l;
    if (!s) {
      fclose(f);
      f = NULL;
      s = sep;
    }
    l = 0;
  }
  if (f)
    fclose(f);
}
/* ---------------------------------------- */
void dumpRom()
{
  unsigned adr, len, i, l;
  char *pM, c, ch[9];

  if (ab != NOTSET) {
    adr = ab;
    len = lb;
  }
  else {
    adr = bb * BLKSIZE;
    len = cb * BLKSIZE;
  }

  l = 0;
  while (len) {
    if (!l) {
      l = readRom(adr, len);
      pM = pMem;
    }

    printf("%04x  ", adr);
    i = 0;
    do {
      c = *pM++;
      printf("%02x ", c);
      ch[i++] = (c < 32) ? '.' : c;
      len--;
      adr++;
      l--;
      if (!l && len) {
        l = readRom(adr, len);
        pM = pMem;
      }
    } while (len && (adr % 8));

    ch[i] = '\0';
    i = (8 - i) * 3 + 1;
    while (i--)
      putchar(' ');
    puts(ch);
    putchar('\n');
  }
  putchar('\n');
}
/* ---------------------------------------- */
unsigned readRom(a, l)
  unsigned a, l;
{
  char *pT;

  pT = TRANS2 + 5;
  *pT++ = (a & 255);
  *pT++ = (a >> 8);
  l = ((l > BLKSIZE) ? BLKSIZE : l)  - 1;
  *pT++ = (l & 255);
  *pT++ = (l >> 8);
  a = pMem;
  *pT++ = (a & 255);
  *pT = (a >> 8);

  call(TRANS);

  return l + 1;
}
/* ---------------------------------------- */
int procArgs(argc, argv)
  int argc;
  char **argv;
{
  int err, i;
  unsigned v;

  err = EOK;
  bb = NOTSET;
  cb = NOTSET;
  sb = NOTSET;
  ab = NOTSET;
  lb = NOTSET;
  dump = FALSE;
  *fn = '\0';
  pFn = NULL;

  do {
    for (i = 1; i < argc; i++) {
      if (!strcmp(argv[i], "-B")) {
        if (i + 1 >= argc) {
          err = prtErr(EOPTVAL, argv[i], NULL);
          break;
        }
        else {
          i++;
          /* pociatocny blok */
          v = atoi(argv[i]);
          if (v > 31) {
            err = prtErr(EBADVAL, argv[i], argv[i - 1]);
            break;
          }
          bb = v;
        }
        continue;
      }
      else if (!strcmp(argv[i], "-C")) {
        if (i + 1 >= argc) {
          err = prtErr(EOPTVAL, argv[i], NULL);
          break;
        }
        else {
          i++;
          /* pocet blokov */
          v = atoi(argv[i]);
          if (v < 1 || v > 32) {
            err = prtErr(EBADVAL, argv[i], argv[i - 1]);
            break;
          }
          cb = v;
        }
        continue;
      }
      else if (!strcmp(argv[i], "-S")) {
        if (i + 1 >= argc) {
          err = prtErr(EOPTVAL, argv[i], NULL);
          break;
        }
        else {
          i++;
          /* pocet blokov na subor */
          v = atoi(argv[i]);
          if (v < 1 || v > 32) {
            err = prtErr(EBADVAL, argv[i], argv[i - 1]);
            break;
          }
          sb = v;
        }
        continue;
      }
      else if (!strcmp(argv[i], "-A")) {
        if (i + 1 >= argc) {
          err = prtErr(EOPTVAL, argv[i], NULL);
          break;
        }
        else {
          i++;
          /* pociatocna adresa */
          if (sscanf(argv[i], "%x", &v) != 1
              || v > 32767) {
            err = prtErr(EBADVAL, argv[i], argv[i - 1]);
            break;
          }
          ab = v;
        }
        continue;
      }
      else if (!strcmp(argv[i], "-L")) {
        if (i + 1 >= argc) {
          err = prtErr(EOPTVAL, argv[i], NULL);
          break;
        }
        else {
          i++;
          /* dlzka citanych dat */
          if (sscanf(argv[i], "%x", &v) != 1
              || v < 1 || v > 32768) {
            err = prtErr(EBADVAL, argv[i], argv[i - 1]);
            break;
          }
          lb = v;
        }
        continue;
      }
      else if (!strcmp(argv[i], "-F")) {
        if (i + 1 >= argc) {
          err = prtErr(EOPTVAL, argv[i], NULL);
          break;
        }
        else {
          i++;
          /* meno suboru */
          pFn = argv[i];
        }
        continue;
      }
      else if (!strcmp(argv[i], "-D")) {
        dump = TRUE;
        continue;
      }
      else {
        err = prtErr(EUNKOPT, argv[i], NULL);
        break;
      }
    }
    if (err != EOK)
      break;

    if ((bb != NOTSET && cb == NOTSET)
        || (bb == NOTSET && cb != NOTSET)) {
      err = prtErr(EMISOPT, "-B alebo -C", NULL);
      break;
    }

    if (bb != NOTSET && bb + cb > 32) {
      sprintf(fn, "%d", cb);
      err = prtErr(EBADVAL, fn, "-C");
      break;
    }

    if (cb != NOTSET && sb != NOTSET && (cb % sb) > 0) {
      sprintf(fn, "%d", sb);
      err = prtErr(EBADVAL, fn, "-S");
      break;
    }

    if ((ab != NOTSET && lb == NOTSET)
        || (ab == NOTSET && lb != NOTSET)) {
      err = prtErr(EMISOPT, "-A alebo -L", NULL);
      break;
    }

    if (ab != NOTSET && ab + lb > 32768) {
      sprintf(fn, "%04x/%04x", lb);
      err = prtErr(EBADVAL, fn, "-A/-L");
      break;
    }

    if (bb == NOTSET && ab == NOTSET) {
      err = prtErr(EMISOPT, "-BCS alebo -AL", NULL);
      break;
    }

    if ((bb != NOTSET || sb != NOTSET) && ab != NOTSET) {
      err = prtErr(EOPTCOMB, "-BCS", "-AL");
      break;
    }

    if (dump && sb != NOTSET) {
      err = prtErr(EOPTCOMB, "-D", "-S");
      break;
    }

    if (dump && pFn) {
      err = prtErr(EOPTCOMB, "-D", "-F");
      break;
    }

    if (!dump && !pFn) {
      err = prtErr(EMISOPT, "-D alebo -F", NULL);
      break;
    }

    if (pFn) {
      if (strlen(pFn) > 12) {
        err = prtErr(EBADFN, pFn, NULL);
        break;
      }
      i = index(pFn, ".");
      if (i != ERROR
          && (!(i > 0 && i <= 8)
              || !*(pFn + i + 1))) {
        err = prtErr(EBADFN, pFn, NULL);
        break;
      }
      if (i != ERROR && i > 0) {
        v = index(pFn + i + 1, ".");
        if (v != ERROR) {
          err = prtErr(EBADFN, pFn, NULL);
          break;
        }
      }
      if (sb == NOTSET)
        strcpy(fn, pFn);
      else {
        if (i != ERROR && i > 0)
          *(pFn + i) = '\0';
        strcpy(fn, pFn);
        nfn = strlen(fn);
        if (nfn > 6) {
          nfn = 6;
          fn[6] = '\0';
        }
        nfn++;
        strcat(fn, "0/");
        if (i != ERROR && i > 0) {
          strcat(fn, ".");
          strcat(fn, pFn + i + 1);
        }
      }
    }

  } while (FALSE);

  return err;
}
/* ---------------------------------------- */
int prtErr(e, p1, p2)
  int e;
  char *p1, *p2;
{
  switch (e) {
    case EALLOC :
      printf("Nepodarilo sa alokovat pamat!");
      break;
    case EOPTVAL :
      printf("Chyba hodnota pre prepinac `%s`!", p1);
      break;
    case EBADVAL :
      printf("Neplatna hodnota `%s` pre prepinac `%s`!", p1, p2);
      break;
    case EUNKOPT :
      printf("Nepodporovany prepinac `%s`!", p1);
      break;
    case EOPTCOMB :
      printf("Nedovolena kombinacia prepinacov `%s` a `%s`!", p1, p2);
      break;
    case EMISOPT :
      printf("Chyba prepinac `%s`!", p1);
      break;
    case EBADFN :
      printf("Chybne meno suboru `%s`!", p1);
      break;
    case EFILCRE :
      printf("Nepodarilo sa vytvorit subor '%s'!", p1);
      break;
    case EFILWRI :
      printf("Zapis do suboru '%s' zlyhal!", p1);
      break;
  }
  printf("\n\n");

  return e;
}
/* ---------------------------------------- */
void version()
{
  printf("\n%s v%d.%d - %s\n\n", PRGNAME,
	 PRGVER / 16, PRGVER % 16, PRGAUTH);
}
/* ---------------------------------------- */
void usage()
{
  printf("Program pre citanie obsahu PMD 85 ROM Modulu.\n");
  printf("Pouzitie: RR [prepinace]\n");
  printf("\nPrepinace:\n");
  printf("  -B DEC   cislo pociatocneho bloku (0-31)\n");
  printf("  -C DEC   pocet citanych blokov (1-32)\n");
  printf("  -S DEC   separatne subory, kazdy DEC blokov\n");
  printf("  -A HEX   pociatocna adresa (0000-7FFF)\n");
  printf("  -L HEX   citana dlzka (0001-8000)\n");
  printf("  -F FN    meno vystupneho suboru\n");
  printf("  -D\t   iba Dump obsahu na obrazovku\n");
  printf("\nBlok ma velkost 1 kB.\n");
  printf("DEC je desiatkove cislo.\n");
  printf("HEX je hexadecimalne cislo.\n");
  printf("Prepinace '-BCS' nemozno kombinovat s '-AL'.\n");
  printf("Iba jeden prepinac '-F' a '-D' mozno pouzit.\n");
  printf("Ak je pouzity prepinac '-S', do zadaneho mena\n");
  printf("suboru sa pridava cislovanie.\n");
}
/* ---------------------------------------- */

